# -*- coding: utf-8 -*-

import os
import sys
import asyncio
import time

from .color import color_str

from .utils import data_pack, data_unpack, sizeof_fmt, terminal_width


class File(object):

    def __init__(self, path):
        self._fd = None
        self._path = path
        self.info = Info(path)


    @property
    def fd(self):
        return self._fd


    @property
    def path(self):
        return self._path


    @property
    def undownload_chucks(self):
        return self.info.find_undownload_chucks()


    def is_init(self):
        return self.info.content_length == 0


    def create_file(self):
        path = self.path
        if os.path.exists(path):
            fd = open(path, 'rb+')
            self._fd = fd
            return fd
        else:
            _dir = os.path.dirname(path)
            if _dir and not os.path.exists(_dir):
                os.makedirs(_dir)
            fd = open(path, 'wb+')
            self._fd = fd
            return fd


    def write(self, data, seek):
        self.fd.seek(seek, 0)
        self.fd.write(data)


    def record_data(self, *datas):
        for data in datas:
            self.info.write(data)


    def close(self):
        self.fd.close()


class Info(object):

    def __init__(self, path):
        self.info_filename = path + '.aget'
        if not os.path.exists(self.info_filename):
            self.initiate_info()
        else:
            self.load_info()


    @property
    def downloaded_chucks(self):
        return self._downloaded_chucks


    @property
    def completed_size(self):
        n = 0
        for begin_point, end_point in self.downloaded_chucks:
            n += end_point - begin_point + 1
        return n


    def initiate_info(self):
        self.content_length = 0
        self._downloaded_chucks = []


    def load_info(self):
        data = open(self.info_filename, 'rb').read()
        N = len(data)
        if not data:
            self.initiate_info()
        else:
            self.content_length = data_unpack(data[:8])
            if N > 8:
                chucks = [data_unpack(data[i:i+8]) for i in range(8, N, 8)]
                self._downloaded_chucks = self.merge_chucks(chucks)
                self._dump_info(self.content_length, self.downloaded_chucks)
            else:
                self._downloaded_chucks = []


    def _dump_info(self, content_length, chucks):
        open(self.info_filename, 'wb').close()   # remove old data
        self.write(content_length)
        for begin_point, end_point in chucks:
            self.write(begin_point)
            self.write(end_point)


    def write(self, data):
        with open(self.info_filename, 'ab') as fd:
            data_byte = data_pack(data)
            fd.write(data_byte)


    def merge_chucks(self, chucks):
        if len(chucks) % 2 != 0:
            chucks = chucks[:-1]

        intervals = [chucks[i:i+2] for i in range(0, len(chucks), 2)]

        downloaded_chucks = _merge_intervals(intervals)
        return downloaded_chucks


    def find_undownload_chucks(self):
        N = self.content_length
        if not self.downloaded_chucks:
            return [[0, N-1]]
        else:
            if len(self.downloaded_chucks) == 1 \
                    and self.downloaded_chucks[0][0] == 0 \
                    and self.downloaded_chucks[0][-1] == N-1:
                return []

        chucks = [[0, 0]] + list(self.downloaded_chucks) + [[N, N]]
        undownload_chucks = _find_gaps(chucks)

        return undownload_chucks


    def remove_aget(self):
        os.remove(self.info_filename)


def _merge_intervals(intervals):
    intervals.sort()

    mt = []

    mt.append(intervals[0])
    for begin_point, end_point in intervals[1:]:
        pre_begin_point, pre_end_point = mt[-1]

        # case 1
        # ----------
        #                -----------
        if pre_end_point + 1 < begin_point:
            n_begin_point = begin_point
            n_end_point = end_point

        # case 2
        # -----------------
        #                  ----------
        #             --------
        #     ------
        else:
            n_begin_point = pre_begin_point
            n_end_point = max(pre_end_point, end_point)
            mt.pop()

        mt.append([n_begin_point, n_end_point])

    return mt


def _find_gaps(intervals):
    if len(intervals) <= 1:
        return []

    gaps = []

    for i, interval in enumerate(intervals[:-1]):
        next_interval = intervals[i+1]
        if interval[1] + 1 < next_interval[0]:
            gaps.append([interval[1] + 1, next_interval[0] - 1])
        else:
            continue

    return gaps


class Shower(object):

    TEMPLATE = ('\n {}: {{}}\n'
                ' {}: {{}} ({{}})\n').format(
                    color_str('File', codes=(1, 92)),
                    color_str('Size', codes=(1, 94))
                )


    def __init__(self, filename,
                 content_length,
                 completed_size,
                 concurrency,
                 chuck_size):

        self.filename = filename
        self.content_length = content_length
        self._completed_size = completed_size
        self._pre_size = completed_size
        self._stop = False
        self._completed_chucks = []


    @property
    def completed_size(self):
        while self._completed_chucks:
            part, begin_point, end_point = self._completed_chucks.pop()
            self._completed_size += (end_point - begin_point + 1)
        return self._completed_size


    @property
    def stop(self):
        return self._stop

    @stop.setter
    def stop(self, value):
        self._stop = True


    async def show(self):
        total_size = sizeof_fmt(self.content_length)
        header = self.TEMPLATE.format(self.filename,
                                      total_size,
                                      self.content_length)
        print(header)

        self._begin_time = time.time()
        while True:
            if self.stop:
                break

            status_line = self._gen_status_line()

            sys.stdout.write(status_line)
            sys.stdout.flush()

            await asyncio.sleep(2)


    def _gen_status_line(self):
        end_time = time.time()

        total_size = sizeof_fmt(self.content_length)
        completed_size = self.completed_size
        download_size = completed_size - self._pre_size
        self._pre_size = completed_size

        speed = download_size / (end_time - self._begin_time)
        speed_str = '{: >7}/s'.format(sizeof_fmt(int(speed)))

        percent = '{:.2f}'.format(completed_size / self.content_length * 100)

        width = terminal_width()
        cs = sizeof_fmt(completed_size)
        pre_width = len('{}/{} {}% {} [] '.format(cs, total_size, percent, speed_str))
        status = '{}/{} {}% {}'.format(
            color_str(cs, codes=(1, 91)),
            color_str(total_size, codes=(1, 92)),
            color_str(percent, codes=(1, 33)),
            color_str(speed_str, codes=(1, 94)))
        process_line_width = width - pre_width
        p = completed_size / self.content_length
        process_line = '>' * int(p * process_line_width) \
                        + ' ' * (process_line_width - int(p * process_line_width))

        status_line = '\r{} [{}] '.format(status, process_line)

        self._begin_time = end_time
        return status_line


    def append_info(self, part, begin_point, end_point):
        self._completed_chucks.append((part, begin_point, end_point))


    def over(self):
        print('\n' + color_str('Over.', codes=(1, 97)))
        self.stop = True



# -*- coding: utf-8 -*-

import os
import functools
import asyncio

from .request import async_request, get_content_length, request_range
from .models import File, Shower
from .utils import make_headers, exit_session
from .color import color_str


async def download(args):
# async def download(method, url,
                   # headers=None,
                   # data=None,
                   # timeout=None,
                   # config=None,
                   # chuck_size=DEFAULT_CHUCK_SIZE,
                   # concurrency=DEFAULT_CONCURRENCY):

    file_obj = File(args.out)

    method = args.method
    url = args.url
    headers = make_headers(args)
    data = args.data
    timeout = args.timeout
    chuck_size = args.chuck_size
    concurrency = args.concurrency

    content_length = await get_content_length(method, url,
                                              headers=headers,
                                              data=data,
                                              timeout=timeout)

    if file_obj.is_init():
        file_obj.record_data(content_length)
    else:
        if file_obj.info.content_length != content_length:
            print(color_str('Conflict content length:', codes=(1, 91)),
                  file_obj.info.content_length,
                  content_length)
            return None

    file_obj.info.content_length = content_length
    file_obj.create_file()

    shower = Shower(args.out,
                    content_length,
                    file_obj.info.completed_size,
                    concurrency,
                    chuck_size)

    show_task = asyncio.ensure_future(shower.show())

    ctrl_queue = asyncio.queues.Queue(maxsize=concurrency)

    part = 1
    # print(list(file_obj.undownload_chucks))
    for begin_point, end_point in file_obj.undownload_chucks:
        point = begin_point
        point_t = 0
        while point <= end_point:
            await ctrl_queue.put(None)
            point_t = min(point + chuck_size, end_point)

            task = asyncio.ensure_future(
                request_range(method, url,
                              point,
                              point_t,
                              ctrl_queue,
                              headers=headers,
                              data=data,
                              timeout=timeout))

            task.add_done_callback(
                functools.partial(save_data, file_obj,
                                  point, point_t,
                                  shower, part))

            if point == point_t:
                break

            point = min(point_t + 1, end_point)
            part += 1

    while ctrl_queue.qsize():
        await asyncio.sleep(1)

    await asyncio.sleep(1)
    file_obj.close()
    file_obj.info.remove_aget()
    exit_session()
    shower.over()
    show_task.cancel()


def save_data(file_obj, begin_point, end_point, shower, part, fut):
    data = fut.result()

    file_obj.write(data, begin_point)
    file_obj.record_data(begin_point, end_point)

    shower.append_info(part, begin_point, end_point)





from http import HTTPStatus

OK_STATUSES = (
    HTTPStatus.OK,
    HTTPStatus.CREATED,
    HTTPStatus.ACCEPTED,
    HTTPStatus.NON_AUTHORITATIVE_INFORMATION,
    HTTPStatus.NO_CONTENT,
    HTTPStatus.RESET_CONTENT,
    HTTPStatus.PARTIAL_CONTENT,
)


# Defines that should never be changed
OneK = 1024
OneM = OneK * OneK
OneG = OneM * OneK
OneT = OneG * OneK

DEFAULT_CHUCK_SIZE = 1 * OneM
DEFAULT_CONCURRENCY = 10



# -*- coding: utf-8 -*-

import os
import configparser
import urllib

from .utils import get_chuck_size


DEFAULT_CONFIG_FILE = os.path.join(
    os.path.dirname(__file__), 'config')

USER_CONFIG_FILE = os.path.join(
    os.path.expanduser('~'), 'config')


def configure(args):
    if os.path.exists(USER_CONFIG_FILE):
        config_file = USER_CONFIG_FILE
    else:
        config_file = DEFAULT_CONFIG_FILE

    assert os.path.exists(config_file), '{} is not existed'.format(config_file)

    config_parser = configparser.ConfigParser(allow_no_value=True)
    config_parser.read(config_file)

    if not args.concurrency:
        concurrency = int(config_parser['basic']['concurrency'])
        args.concurrency = concurrency

    if not args.chuck_size:
        chuck_size = get_chuck_size(config_parser['basic']['chuck_size'])
        args.chuck_size = chuck_size
    else:
        chuck_size = get_chuck_size(args.chuck_size)
        args.chuck_size = chuck_size

    if not args.header:
        args.header = ['User-Agent: ' + config_parser['http']['user-agent'] or 'aget']

    if config_parser['basic']['quiet'] == 'true':
        args.quiet = True

    if args.data:
        args.method = 'POST'

    if not args.out:
        url_parser = urllib.parse.urlparse(args.url)
        path = url_parser.path.strip('/')
        if not path:
            raise IOError('Can not get filename')
        else:
            args.out = path.split('/')[-1]



# -*- coding: utf-8 -*-

import asyncio
import mugen

from .common import OK_STATUSES
from .exceptions import ContentLengthError


async def async_request(method, url, **kwargs):
    while True:
        try:
            resp = await mugen.request(method, url, **kwargs)
            if resp.status_code in OK_STATUSES:
                return resp
        except Exception:
            await asyncio.sleep(1)


async def request_range(method, url, start, end, ctrl_queue, **kwargs):
    headers = {'Range': 'bytes={}-{}'.format(start, end)}
    headers.update(kwargs.get('headers') or {})
    kwargs['headers'] = headers

    resp = await async_request(method, url, **kwargs)
    await ctrl_queue.get()
    return resp.content


async def get_content_length(method, url, **kwargs):
    # get size from HEAD
    if method.lower() == 'get':
        _method = 'HEAD'

        headers = kwargs.get('headers')
        kwargs['headers'] = headers
        resp = await async_request(_method, url, **kwargs)
        if resp.headers.get('Content-Length'):
            size = int(resp.headers.get('Content-Length'))
            return size

    # get size from range
    headers = kwargs['headers'] or {}
    headers['Range'] = {'bytes=0-1'}
    kwargs['headers'] = headers
    resp = await async_request(method, url, **kwargs)
    if resp.headers.get('Content-Range'):
        size = int(resp.headers['Content-Range'].split('/')[-1])
        return size

    raise ContentLengthError('Server does not support partially downloaded')



# -*- coding: utf-8 -*-

import sys
import os
import struct

import mugen

from .color import color_str
from .common import (
    OneK,
    OneM,
    OneG,
    OneT,
    DEFAULT_CHUCK_SIZE
)


STRUCT_FORMAT = 'Q'   # unsigned long long


def make_headers(args):
    headers = {}
    for header in args.header:
        k, v = header.split(': ', 1)
        headers[k] = v
    return headers


def data_pack(num):
    return struct.pack(STRUCT_FORMAT, num)


def data_unpack(chuck):
    return struct.unpack(STRUCT_FORMAT, chuck)[0]


def exit_session():
    mugen.session().close()


def sizeof_fmt(num):
    for x in ['B','K','M','G']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'T')


def get_chuck_size(chuck_size_str):
    size = chuck_size_str.upper()
    if size.isdigit():
        return int(size)
    elif size.endswith('K'):
        s = int(size[:-1]) * OneK
        return s
    elif size.endswith('M'):
        s = int(size[:-1]) * OneM
        return s
    else:
        return DEFAULT_CHUCK_SIZE


def terminal_width():
    return int(os.popen('tput cols').read())


def assert_completed_file(args):
    path = args.out
    info_path = args.out + '.aget'
    if os.path.exists(path) and not os.path.exists(info_path):
        print(color_str(path, codes=(1, 91)), 'is existed')
        sys.exit()



# -*- coding: utf-8 -*-

import sys
import signal
import asyncio
import argparse

import mugen

from .download import download
from .color import color_str
from .configure import configure
from .utils import assert_completed_file


def sigal_handler(signum, frame):
    print(color_str("  !! Signal:", codes=(1, 91)), signum)
    mugen.session().close()

    sys.exit(1)


def handle_signal():
    signal.signal(signal.SIGQUIT, sigal_handler)
    signal.signal(signal.SIGINT,  sigal_handler)
    signal.signal(signal.SIGTERM, sigal_handler)


def parse_arguments(argv):
    p = argparse.ArgumentParser(description='')

    p.add_argument('url', type=str, help='')
    p.add_argument('-H', '--header', action='append',
                   default=[], help='header')
    p.add_argument('-X', '--method', action='store',
                   default='GET', type=str, help='')
    p.add_argument('-s', '--concurrency', action='store',
                   default=None, type=int, help='')
    p.add_argument('-k', '--chuck_size', action='store',
                   default=None, type=str, help='')
    p.add_argument('-d', '--data', action='store',
                   default=None, type=str, help='')
    p.add_argument('-t', '--timeout', action='store',
                   default=None, type=int, help='')
    p.add_argument('-q', '--quiet', action='store_true',
                   default=False, help='')
    p.add_argument('-o', '--out', action='store',
                   default=None, type=str, help='')

    args = p.parse_args(argv)

    return args


def main():
    handle_signal()

    argv = sys.argv[1:]

    args = parse_arguments(argv)

    configure(args)

    assert_completed_file(args)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(download(args))



# -*- coding: utf-8 -*-

COLOR_TEMPLATE = '\x1b[{}m{}\x1b[0m'

#
# http://misc.flogisoft.com/bash/tip_colors_and_formatting
#
# formatting text
#
# 1    # Bold/Bright
# 2    # Dim
# 4    # Underlined
# 5    # Blink
# 7    # Reverse (invert the foreground and background colors)
# 8    # Hidden (usefull for passwords)
#
#
# 8/16 Colors
#
# Foreground (text)
# 39   # Default foreground color    Default Default
# 30   # Black   Default Black
# 31   # Red Default Red
# 32   # Green   Default Green
# 33   # Yellow  Default Yellow
# 34   # Blue    Default Blue
# 35   # Magenta Default Magenta
# 36   # Cyan    Default Cyan
# 37   # Light gray  Default Light gray
# 90   # Dark gray   Default Dark gray
# 91   # Light red   Default Light red
# 92   # Light green Default Light green
# 93   # Light yellow    Default Light yellow
# 94   # Light blue  Default Light blue
# 95   # Light magenta   Default Light magenta
# 96   # Light cyan  Default Light cyan
# 97   # White   Default White
#
# Background
# 49   # Default background color    Default Default
# 40   # Black   Default Black
# 41   # Red Default Red
# 42   # Green   Default Green
# 43   # Yellow  Default Yellow
# 44   # Blue    Default Blue
# 45   # Magenta Default Magenta
# 46   # Cyan    Default Cyan
# 47   # Light gray  Default Light gray
# 100  # Dark gray   Default Dark gray
# 101  # Light red   Default Light red
# 102  # Light green Default Light green
# 103  # Light yellow    Default Light yellow
# 104  # Light blue  Default Light blue
# 105  # Light magenta   Default Light magenta
# 106  # Light cyan  Default Light cyan107 White


def color_str(msg, codes=None):
    if not msg:
        return msg

    if codes:
        return COLOR_TEMPLATE.format(';'.join([str(e) for e in codes]), msg)
    else:
        return msg






import sys


try:
    from setuptools import setup
except ImportError:
    # Use distutils.core as a fallback.
    # We won't be able to build the Wheel file on Windows.
    from distutils.core import setup

if sys.version_info < (3, 5, 0):
    raise RuntimeError("aget requires Python 3.5.0+")

version = "0.1.10"

requires = ["mugen"]

setup(
    name="aget",
    version=version,
    author="PeterDing",
    author_email="dfhasyt@gmail.com",
    license="Apache 2.0",

    description="",
    url="http://github.com/PeterDing/aget",

    install_requires=requires,

    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],

    packages=["aget"],

    scripts=['bin/aget'],

    include_package_data=True
)




from aget.models import _merge_intervals, _find_gaps


intervals = [
    (0, 0),
    (2, 10),
    (3, 8),
    (3, 4),
    (9, 20),
    (30, 40),
    (40, 40)
]


intervals = _merge_intervals(intervals)
print(intervals)

gaps = _find_gaps(intervals)
print(gaps)
